#!/usr/bin/env python

'''

STATE
-----

State is a list of cells, each cell is
  -1: empty
   0: occupied by player 0
   1: occupied by player 1
'''

CELL_EMPTY = -1

# float factor to break ties in order to ensure win is delayed/advanced in
# favour of eval_player.
TIE_BREAK_WIN_EARLY = .001

__docformat__ = 'restructuredtext'
__version__ = '$Id: $'

import random

class Turn(object):
    def apply(self, layout, state):
        raise NotImplementedError()

class TwistRing(Turn):
    def __init__(self, ring, twist):
        self.ring = ring
        self.twist = twist

    def apply(self, layout, state):
        assert layout.ring_directions[self.ring] == self.twist
        n = self.twist
        rs = layout.ring_starts[self.ring]
        rc = layout.ring_counts[self.ring]
        n = n % rc
        return (state[:rs] + 
                state[rs + n:rs + rc] + state[rs:rs + n] +
                state[rs + rc:])

    def __repr__(self):
        return 'TwistRing(ring %d by %d)' % (self.ring, self.twist)

class MovePiece(Turn):
    def __init__(self, cellfrom, cellto):
        self.cellfrom = cellfrom
        self.cellto = cellto

    def apply(self, layout, state):
        assert layout.can_move(state, self.cellfrom, self.cellto)
        state = list(state)
        state[self.cellto] = state[self.cellfrom]
        state[self.cellfrom] = CELL_EMPTY
        return tuple(state)

    def __repr__(self):
        return 'MovePiece(%d to %d)' % (self.cellfrom, self.cellto)

class BoardLayout(object):
    def __init__(self, divisions):
        nrings = len(divisions)
        ring_counts = divisions[:]
        for i in range(1, len(ring_counts)):
            ring_counts[i] *= ring_counts[i - 1]
        ring_starts = [0] + ring_counts[:-1]
        for i in range(1, len(ring_starts)):
            ring_starts[i] += ring_starts[i - 1]

        ncells = sum(ring_counts)
        cell_neighbours = [[] for i in range(ncells)]

        self.ring_directions = [(i % 2) * 2 - 1 for i in range(nrings)]

        # same ring neighbours 
        for (ring_count, ring_start) in zip(ring_counts, ring_starts):
            for i in range(ring_start + 1, ring_start + ring_count):
                cell_neighbours[i].append(i - 1)
            if ring_count != 1:
                cell_neighbours[ring_start].append(ring_start + ring_count - 1)
            for i in range(ring_start, ring_start + ring_count - 1):
                cell_neighbours[i].append(i + 1)
            if ring_count != 1:
                cell_neighbours[ring_start + ring_count - 1].append(ring_start)

        # adjacent ring neighbours
        for ring in range(1, nrings):
            for outer in range(ring_counts[ring]):
                inner = outer // divisions[ring]
                outer_ind = ring_starts[ring] + outer
                inner_ind = ring_starts[ring - 1] + inner
                cell_neighbours[outer_ind].append(inner_ind)
                cell_neighbours[inner_ind].append(outer_ind)

        # board sides
        self.cell_owners = []
        for ring in range(nrings):
            for i in range(ring_counts[ring]):
                owner = i // (ring_counts[ring] // 2)
                self.cell_owners.append(owner)

        self.cell_rings = []
        for ring in range(nrings):
            for i in range(ring_counts[ring]):
                self.cell_rings.append(ring)

        # distances to halfway mark along current ring twist
        self.cell_twist_distances = []
        for ring in range(nrings):
            halfway = ring_counts[ring] // 2
            for i in range(ring_counts[ring]):
                if self.ring_directions[ring] < 0:
                    self.cell_twist_distances.append(
                        (ring_counts[ring] - i - 1) % halfway + 1)
                else:
                    self.cell_twist_distances.append(i % halfway + 1)

        self.ncells = ncells
        self.nrings = nrings
        self.divisions = divisions
        self.ring_counts = ring_counts
        self.ring_starts = ring_starts
        self.ringstate_start = ncells
        self.cell_neighbours = cell_neighbours
        self.players = 2

        self.initial_state = [CELL_EMPTY for i in range(self.ncells)] 
        goal = 0
        for ring in range(0, self.nrings):
            for i in range(self.ring_starts[ring] + (ring % 2), 
                           self.ring_starts[ring] + self.ring_counts[ring], 2):
                self.initial_state[i] = 1 - self.cell_owners[i]
                goal += 1 
        self.goal_score = goal // 2

    def initial(self):
        return tuple(self.initial_state)

    def get_possible_turns(self, state, player):
        # Ring twists
        for ring, twist in enumerate(self.ring_directions):
            yield TwistRing(ring, twist)

        # Move pieces
        for fromcell, piece in enumerate(state):
            if piece == player:
                for tocell in self.cell_neighbours[fromcell]:
                    if state[tocell] == CELL_EMPTY:
                        yield MovePiece(fromcell, tocell)

    def get_possible_turn_states(self, state, player):
        for turn in self.get_possible_turns(state, player):
            newstate = turn.apply(self, state)
            if newstate != state:
                yield turn, newstate

    def best_turn(self, state, player, depth):
        import time
        start_time = time.time()
        best = []
        best_score = -1000
        self.trans_table = {}
        turns = list(self.get_possible_turn_states(state, player))
        for idepth in range(depth):
            for turn, newstate in turns:
                turn.score = self.alphabeta(newstate, idepth, 
                                            player, -10000, 10000)
            turns.sort(key=lambda o: o[0].score, reverse=True)
            #print 'idepth = %d' % idepth
            #for turn, _ in turns:
            #    print '  %r = %f' % (turn, turn.score)
        #print 'time: %.02f' % (time.time() - start_time)
        #print 'ttable size: %d' % (len(self.trans_table))
        if depth == 1:
            return random.choice(turns[:4])[0]
        elif depth == 2:
            return random.choice(turns[:2])[0]
        else:
            return turns[0][0]

    # transposition table maps (state, player) -> (depth, score, limit)
    # limit is -1 if result is beta (lower limit)
    #           0 if result is exact
    #           1 if result is alpha (upper limit)
    trans_table = {}

    def alphabeta(self, state, depth, player, alpha, beta):
        trans_depth, trans_score, trans_limit = \
            self.trans_table.get((state, player), (-1, 0, 0))
        if trans_depth >= depth and False:
            # Tricksy validation of stored transposition value.  This
            # page was helpful: 
            # http://www.seanet.com/~brucemo/topics/hashing.htm
            if trans_limit == -1:
                if trans_score >= beta:
                    return beta
            elif trans_limit == 0:
                return trans_score
            else: # 1
                if trans_score <= alpha:
                    return alpha

        terminal = self.is_won(state)
        if depth == 0 or terminal:
            score = self.ai_eval(state, player)
            if terminal:
                # Tie-breaker: win earlier in favour of player
                score -= depth * TIE_BREAK_WIN_EARLY
            self.trans_table[(state, player)] = (score, depth, 0)
            return score
        trans_limit = 1
        for turn, newstate in self.get_possible_turn_states(state, player):
            val = -self.alphabeta(newstate, depth - 1, 1 - player, 
                                  -beta, -alpha)
            if val >= beta:
                self.trans_table[(state, player)] = (beta, depth, -1)
                return beta
            if val > alpha:
                trans_limit = 0
                alpha = val
        self.trans_table[(state, player)] = (alpha, depth, trans_limit)
        return alpha

    def ai_eval(self, state, player, dump=False):
        player_score = 0
        player_dist = 0
        opponent_score = 0
        opponent_dist = 0

        for piece, owner, twist_dist in \
             zip(state, self.cell_owners, self.cell_twist_distances):
            if piece == -1:
                continue
            if piece == owner:
                # In goal area
                if piece == player:
                    player_score += 1
                else:
                    opponent_score += 1
            else:
                # Not in goal area
                if piece == player:
                    player_dist += twist_dist
                else:
                    opponent_dist += twist_dist

        # Should favour winning over drawing, and drawing over losing
        if player_score == self.goal_score:
            player_score = 1000
        if opponent_score == self.goal_score:
            opponent_score = 1000

        aggregate = (player_score * 6 + \
                     opponent_score * -3 + \
                     player_dist * -2 + \
                     opponent_dist * 1)
        if dump:
            print zip(state, self.cell_owners, self.cell_twist_distances)
            print 'player score', player_score, 'dist', player_dist
            print 'opponent score', opponent_score, 'dist', opponent_dist
            print 'aggregate', aggregate
            
        return aggregate

    def is_won(self, state):
        return self.goal_score in self.get_player_scores(state)

    def get_winner(self, state):
        assert self.is_won(state)
        return self.get_player_scores(state).index(self.goal_score)
                    
    def can_move(self, state, cellfrom, cellto):
        if cellfrom == cellto:
            return False
        if cellto not in self.cell_neighbours[cellfrom]:
            return False
        if state[cellto] != CELL_EMPTY:
            return False
        return True

    def get_player_scores(self, state):
        scores = [0, 0]
        for cell, player in enumerate(state):
            if player == self.cell_owners[cell]:
                scores[player] += 1
        return scores

