#!/usr/bin/env python

'''
'''

__docformat__ = 'restructuredtext'
__version__ = '$Id: $'

from math import sin, cos, pi, atan2, sqrt, floor

from pyglet.gl import *
from pyglet.gl import gl_info
from pyglet import image

import euclid
import susan
from res import getres

INNER_DISC_RADIUS = .5
INNER_DISC_SLICES = 80
INNER_DISC_RAISE_Z = .08

RING_RADIUS = 1.
RING_RADIUS_SEPARATION = 0.01
RING_HEIGHT = .5
RING_BEVEL = 0.03
RING_SLICES = 400
RING_TEXTURE_RADIUS = INNER_DISC_RADIUS + \
                      3 * (RING_RADIUS + RING_RADIUS_SEPARATION) + \
                      RING_HEIGHT + .1
ETCH_Z = 0.02

PIECE_SLICES = 40
PIECE_RADIUS = 0.3
PIECE_HEIGHT = 0.1
PIECE_BEVEL = 0.03
PIECE_DRAG_HEIGHT = 0

# lathe around z to produce torus.  slice contains (rad, z) points
def draw_lathe(point_fn, slices, texr=RING_TEXTURE_RADIUS):
    vertices = []
    normals = []
    tex_coords = []

    a = 0
    incr = 2 * pi / slices
    lastslice = [(r, 0, z, r) for (r, z) in point_fn(slices)]
    for i in range(slices + 1):
        ca, sa = cos(a), sin(a)
        newslice = [(ca * r, sa * r, z, r) for (r, z) in point_fn(i)]
        lastp = None
        for ((x1, y1, z1, r1), (x2, y2, z2, r2)) in zip(lastslice, newslice):
            p = euclid.Point3(x1, y1, z1)
            if lastp:
                t1 = euclid.Vector3(x2 - x1, y2 - y1, z2 - z1)
                t2 = p - lastp
                n = t1.cross(t2)
                normal = (n.x, n.y, n.z)
            else:
                normal = (0, 0, -1)
            lastp = p
            r1 = (r1 + abs(z1)) / texr / 2
            r2 = (r2 + abs(z2)) / texr / 2
            normals.extend(normal)
            tex_coords.extend((ca * r2 + .5, sa * r2 + .5))
            vertices.extend((x2, y2, z2))
            normals.extend(normal)
            tex_coords.extend((ca * r1 + .5, sa * r1 + .5))
            vertices.extend((x1, y1, z1))
        lastslice = newslice
        a += incr

    vertices = (GLfloat * len(vertices))(*vertices)
    normals = (GLfloat * len(normals))(*normals)
    tex_coords = (GLfloat * len(tex_coords))(*tex_coords)
    count = 2 * len(lastslice)
    counts = [count for i in range(slices + 1)]
    counts = (GLint * len(counts))(*counts)
    firsts = range(0, (slices + 1) * count, count)
    firsts = (GLint * len(firsts))(*firsts)
    glPushAttrib(GL_ENABLE_BIT)
    glEnable(GL_NORMALIZE)
    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT)
    glEnableClientState(GL_VERTEX_ARRAY)
    glEnableClientState(GL_NORMAL_ARRAY)
    glEnableClientState(GL_TEXTURE_COORD_ARRAY)
    glVertexPointer(3, GL_FLOAT, 0, vertices)
    glNormalPointer(GL_FLOAT, 0, normals)
    glTexCoordPointer(2, GL_FLOAT, 0, tex_coords)
    if gl_info.have_version(1, 4):
        glMultiDrawArrays(GL_QUAD_STRIP, firsts, counts, slices + 1)
    else:
        for first in firsts:
            glDrawArrays(GL_QUAD_STRIP, first, count)
    glPopClientAttrib()
    glPopAttrib()

class Material(object):
    def __init__(self, 
                 front_ambient,
                 front_diffuse,
                 front_specular,
                 front_shininess):
        self.front_ambient = (GLfloat * 4)(*front_ambient)
        self.front_diffuse = (GLfloat * 4)(*front_diffuse)
        self.front_specular = (GLfloat * 4)(*front_specular)
        self.front_shininess = front_shininess

    def apply(self, alpha=None):
        glMaterialfv(GL_FRONT, GL_AMBIENT, self.front_ambient)
        glMaterialfv(GL_FRONT, GL_SPECULAR, self.front_specular)
        glMaterialfv(GL_FRONT, GL_DIFFUSE, self.front_diffuse)
        glMaterialf(GL_FRONT, GL_SHININESS, self.front_shininess)

class BoardView(object):
    player_materials = [
        Material((0, 0, .6, 1),
                 (0, 0, 0, .5),
                 (1, 1, 1, 1),
                 3,),
        Material((.6, 0, 0, 1),
                 (0, 0, 0, .5),
                 (1, 1, 1, 1),
                 3,),
    ]

    board_material = Material(
        (1, 1, 1, 1),
        (1, 1, 1, 1),
        (0, 0, 0, 1),
        0,)

    glow_material = board_material

    # (ring, divs) -> list name
    ring_lists = {}
    inner_disc_list = None
    piece_list = None
    glow_list = None
    glow_texture = None

    def __init__(self, layout, texture):
        cls = self.__class__

        self.board_texture = image.load(texture).texture

        self.layout = layout
        self.ring_angles = [0. for i in range(layout.nrings)]
        self.perm_ring_angles = self.ring_angles[:]
        list1 = glGenLists(layout.nrings)
        self.ring_lists = range(list1, list1 + layout.nrings)

        # Inner disc
        if not cls.inner_disc_list:
            def point_fn(i):
                return [(0, 0),
                        (INNER_DISC_RADIUS/4, 0),
                        (INNER_DISC_RADIUS/2, 0),
                        (INNER_DISC_RADIUS*3/4, 0),
                        (INNER_DISC_RADIUS - RING_BEVEL, 0),
                        (INNER_DISC_RADIUS, -RING_BEVEL),
                        (INNER_DISC_RADIUS, -RING_HEIGHT + RING_BEVEL),
                        (INNER_DISC_RADIUS - RING_BEVEL, -RING_HEIGHT),
                       ]
            cls.inner_disc_list = glGenLists(1)
            glNewList(cls.inner_disc_list, GL_COMPILE)
            draw_lathe(point_fn, INNER_DISC_SLICES)
            glEndList()

        # Rings
        self.ring_radii = []
        inner_radius = INNER_DISC_RADIUS + RING_RADIUS_SEPARATION
        outer_radius = inner_radius + RING_RADIUS
        for ring_i, divisions in enumerate(self.layout.ring_counts):
            if (ring_i, divisions) not in cls.ring_lists:
                slices = RING_SLICES * (ring_i + 1) 
                div_slices = slices / divisions
                def point_fn(i):
                    topz = 0
                    if i % div_slices == 0:
                        topz = -ETCH_Z
                    return [(inner_radius, -RING_HEIGHT),
                            (inner_radius, topz - RING_BEVEL),
                            (inner_radius + RING_BEVEL, topz), 
                            (outer_radius - RING_BEVEL, topz),
                            (outer_radius, topz - RING_BEVEL),
                            (outer_radius, -RING_HEIGHT + RING_BEVEL),
                            (outer_radius - RING_BEVEL, -RING_HEIGHT),
                           ]

                glNewList(self.ring_lists[ring_i], GL_COMPILE)
                draw_lathe(point_fn, slices)
                glEndList()
                cls.ring_lists[ring_i, divisions] = self.ring_lists[ring_i]
            else:
                self.ring_lists[ring_i] = cls.ring_lists[ring_i, divisions]

            self.ring_radii.append(inner_radius)
            inner_radius += RING_RADIUS + RING_RADIUS_SEPARATION
            outer_radius += RING_RADIUS + RING_RADIUS_SEPARATION

        # Piece
        if not cls.piece_list:
            def point_fn(i):
                return [(0, PIECE_HEIGHT),
                        (PIECE_RADIUS/2, PIECE_HEIGHT),
                        (PIECE_RADIUS - PIECE_BEVEL, PIECE_HEIGHT),
                        (PIECE_RADIUS, PIECE_HEIGHT - PIECE_BEVEL),
                        (PIECE_RADIUS, PIECE_BEVEL),
                        (PIECE_RADIUS - PIECE_BEVEL, 0),
                       ]
            cls.piece_list = glGenLists(1)
            glNewList(cls.piece_list, GL_COMPILE)
            glEnable(GL_BLEND)
            glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA)
            glCullFace(GL_FRONT)
            draw_lathe(point_fn, PIECE_SLICES)
            glCullFace(GL_BACK)
            draw_lathe(point_fn, PIECE_SLICES)
            glDisable(GL_BLEND)
            glEndList()

        if cls.glow_list is None:
            cls.glow_texture = image.load(getres('glow.png')).texture
            cls.glow_list = glGenLists(1)
            r = PIECE_RADIUS * 2
            glNewList(self.glow_list, GL_COMPILE)
            glEnable(GL_BLEND)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE)
            self.glow_material.apply()
            self.glow_texture.blit(-r, -r, z=.01, width=r*2, height=r*2)
            glEndList()

        self.drag_cell = None
        self.drag_piece_pos = None
        self.drag_piece_valid = True
        self.inner_disc_z = 0

    def draw_board(self):
        glEnable(self.board_texture.target)
        glBindTexture(self.board_texture.target, self.board_texture.id)
        self.board_material.apply()
        glPushMatrix()
        glTranslatef(0, 0, self.inner_disc_z)
        glCallList(self.inner_disc_list)
        glPopMatrix()
        for ring, (list, perm_angle, angle) in enumerate(
                zip(self.ring_lists, self.perm_ring_angles, self.ring_angles)):
            glPushMatrix()
            glRotatef(angle + perm_angle, 0, 0, 1)
            glCallList(list)
            glPopMatrix()
        glDisable(self.board_texture.target)

    def draw(self, state):
        glEnable(GL_LIGHTING)
        glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, 1)
        glEnable(GL_LIGHT0)
        glLightfv(GL_LIGHT0, GL_POSITION, (GLfloat * 4)(1,1,-4, 0))
        glLightfv(GL_LIGHT0, GL_DIFFUSE, (GLfloat * 4)(.5, .5, .5, 1))
        glLightfv(GL_LIGHT0, GL_SPECULAR, (GLfloat * 4)(1, 1, 1, 1))
        glEnable(GL_LIGHT1)
        glLightfv(GL_LIGHT1, GL_POSITION, (GLfloat * 4)(0, -1, -3, 0))
        glLightfv(GL_LIGHT1, GL_DIFFUSE, (GLfloat * 4)(.5, .5, .5, 1))
        glLightfv(GL_LIGHT1, GL_SPECULAR, (GLfloat * 4)(1, 1, 1, 1))

        self.draw_board()

        self.piece_positions = []
        ring = 0
        for cell, player in enumerate(state):
            if ring < self.layout.nrings - 1 and \
               cell == self.layout.ring_starts[ring + 1]:
                ring += 1
            if player != susan.CELL_EMPTY:
                pos = self.get_cell_pos(cell)
                z = 0
                glPushMatrix()
                glRotatef(self.ring_angles[ring], 0, 0, 1)
                if cell == self.drag_cell:
                    glPushAttrib(GL_LIGHTING_BIT)
                    if not self.drag_piece_valid:
                        glLightModelfv(GL_LIGHT_MODEL_AMBIENT, 
                                   (GLfloat * 4)(0, 0, 0, 1))
                    glTranslatef(self.drag_piece_pos.x, 
                                 self.drag_piece_pos.y, PIECE_DRAG_HEIGHT)
                    self.player_materials[player].apply()
                    glCallList(self.piece_list)
                    glPopAttrib()
                else:
                    glTranslatef(pos.x, pos.y, z)
                    self.piece_positions.append((cell, pos.x, pos.y, z))
                    if self.layout.cell_owners[cell] == player:
                        glCallList(self.glow_list)
                    self.player_materials[player].apply()
                    glCallList(self.piece_list)
                glPopMatrix()

    def intersect_ray_board(self, ray):
        plane = euclid.Plane(euclid.Point3(0., 0., 0.),
                             euclid.Vector3(0., 0., 1.))

        return plane.intersect(ray)

    def get_intersection_angle(self, point):
        angle = atan2(point.y, point.x) 
        if angle < 0:
            angle = 2 * pi + angle
        return angle * 180 / pi

    def get_twist_for_angle(self, ring, angle):
        angle = -angle
        if angle:
            dir = angle / abs(angle)
            return int(angle * self.layout.ring_counts[ring] / 360. + dir * .2)
        return 0

    def get_angle_for_twist(self, ring, twist):
        return -twist * 360. / self.layout.ring_counts[ring]

    def get_ring_angle_limits(self, ring):
        min = 0.
        max = 360. / self.layout.ring_counts[ring]
        if self.layout.ring_directions[ring] == 1:
            min = -max
            max = 0.
        return min, max

    def get_cell_pos(self, cell):
        ring = self.layout.cell_rings[cell]
        radius = self.ring_radii[ring] + RING_RADIUS / 2
        angle = 2 * pi / self.layout.ring_counts[ring] * \
                    (cell - self.layout.ring_starts[ring] + .5)
        x = radius * cos(angle)
        y = radius * sin(angle)
        return euclid.Point3(x, y, 0)

    def get_intersection_radius(self, point):
        return sqrt(point.x * point.x + point.y * point.y)

    def get_intersection_ring(self, point):
        radius = self.get_intersection_radius(point)
        ring = 0
        for ring, ring_rad in enumerate(self.ring_radii):
            if ring_rad < radius < (ring_rad + RING_RADIUS):
                return ring

    def get_intersection_center(self, point):
        radius = self.get_intersection_radius(point)
        return radius < self.ring_radii[0]
        
    def get_intersection_cell(self, point):
        ring = self.get_intersection_ring(point)
        if ring is None:
            return None

        angle = self.get_intersection_angle(point)
        div = int(angle / 360. * self.layout.ring_counts[ring])
        div %= self.layout.ring_counts[ring]
        return self.layout.ring_starts[ring] + div

    def get_intersection_piece(self, ray):
        '''return (cell, ref_point) or None'''

        for cell, x, y, z in self.piece_positions:
            plane = euclid.Plane(euclid.Point3(0., 0., z),
                                 euclid.Vector3(0., 0., 1.))
            point = plane.intersect(ray)
            dist2 = (point.x - x) ** 2 + (point.y - y) ** 2
            if dist2 < PIECE_RADIUS ** 2:
                return cell, euclid.Point3(point.x - x, point.y - y, 0)
        return None, None

    def save_ring_angles(self):
        # Round to nearest click first
        for i, a in enumerate(self.ring_angles):
            incr = 360. / self.layout.ring_counts[i]
            self.ring_angles[i] = floor(a / incr + .5) * incr

        self.perm_ring_angles = \
            [a + b for (a, b) in zip(self.ring_angles, self.perm_ring_angles)]

    def reset_angles(self):
        self.ring_angles = [0. for i in self.ring_angles]
        self.perm_ring_angles = [0. for i in self.perm_ring_angles]
